!> @file libmtime.f90
!!
!! @brief Providing the Fortran language bindings for libmtime
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!! @defgroup FortranBindings libmtime Fortran languange bindings
!! @{
!!
!___________________________________________________________________________________________________________
!>
!! @brief Provides the calendar to the users, abstracting the different calendars available.
!!
!! @details
!!
!! Three calendar types are provided:
!!
!!   - a proleptic Gregorian calendar
!!   - a calendar with 365 days per year without leap years
!!   - a calendar with 360 days per year and each month having 30 days
!!
!! To use a specific calendar a call to setCalendar with the respective selector must be
!! done. The implementation is based on a singleton concept meaning that only one calendar
!! can be active at a time. To release a calendar a call to resetCalendar has to be done.
!!
!___________________________________________________________________________________________________________
module mtime_calendar
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_null_char, c_ptr, c_associated
  !
  implicit none
  !
  private
  !   
  enum, bind(c)
    enumerator :: calendar_not_set    = 0   ! calendar is not defined yet
    enumerator :: proleptic_gregorian = 1   ! proleptic Gregorian calendar
    enumerator :: year_of_365_days    = 2   ! 365 day year without leap years
    enumerator :: year_of_360_days    = 3   ! 360 day year with 30 day months
  end enum
  !
  public :: max_calendar_str_len
  public :: calendar_not_set, proleptic_gregorian, year_of_365_days, year_of_360_days
  public :: setCalendar
  public :: resetCalendar
  public :: calendarType
  public :: calendarToString
  !
  !> provides a string length for toString 
  integer, parameter :: max_calendar_str_len = 32
  !
  !> @cond DOXYGEN_IGNORE_THIS
  interface
    !
    subroutine setCalendar(ct) bind(c, name='initCalendar') !TESTED-OK
      import :: c_int
      integer(c_int), value :: ct
    end subroutine setCalendar
    !
    subroutine resetCalendar() bind(c, name='freeCalendar') !TESTED-OK
    end subroutine resetCalendar
    !
    function calendarType() bind(c, name='getCalendarType') !TESTED-OK
      import :: c_int
      integer(c_int) :: calendarType
    end function calendarType
    !
    function my_calendartostring(calendar) result(c_pointer) bind(c, name='calendarToString') !TESTED-OK
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: calendar
    end function my_calendartostring
    !
  end interface
  !> @endcond DOXYGEN_IGNORE_THIS
  !
contains
  !
#ifdef DOXYGEN_DOCUMENTATION_ONLY
  !> 
  !! @brief Initialize a new calendar.
  !!
  !! setCalendar is done at the very begining to select one of the
  !! provided calendar libraries. It intializes the calendar to one of:
  !!
  !! - proleptic_gregorian
  !! - year_of_365_days
  !! - year_of_360_days
  !! 
  !! The calendar type and hence it's behaviour (Calendar to Julian
  !! conversion and vice versa) is fixed for the lifetime of the
  !! selected calendar.  Attempts to change the calendar type on the
  !! fly is discouraged. The lib has built-in checks to reject
  !! change attempts at run time.  However, a calendar can be
  !! "re-initialized" after calling resetCalendar(), but this is not
  !! advised. 
  !!
  !! MANTRA: Know what you are doing before you do it and do it
  !! right the first time.
  !!
  !! @param[in]  ct      the calendar type
  subroutine setCalendar(ct) bind(c, name='initCalendar') !TESTED-OK
    integer(c_int), value :: ct
  end subroutine setCalendar
  !>
  !! @brief called to discard the selected calendar type 
  subroutine resetCalendar() bind(c, name='freeCalendar') !TESTED-OK
  end subroutine resetCalendar
  !>
  !! @brief to get an idea, which calendar type is selected
  !!
  !! @return an integer denoting the calendar in use
  function calendarType() bind(c, name='getCalendarType') !TESTED-OK
    integer(c_int) :: calendarType
  end function calendarType
  !
#endif
  !>
  !! @brief convert the calendar identifier into a human readable string
  !!
  !! @param[out]  string      the calendar type verbose
  !!
  subroutine calendarToString(string, errno) !TESTED-OK
    character(len=max_calendar_str_len), intent(out) :: string
    integer :: i
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_calendartostring(string)
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 0 * 100 + 4
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine calendarToString
  !
end module mtime_calendar
!>
!! @brief Julian Day Calendar and some operations supported on julian dates.
!!
!! @details
!___________________________________________________________________________________________________________
module mtime_julianday
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_char, c_null_char, c_ptr, c_loc, &
                                       & c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_julianday_str_len
  public :: julianday
  public :: newJulianday
  public :: deallocateJulianday
  public :: juliandayToString
  !
  !> provides a string length for toString 
  integer, parameter :: max_julianday_str_len = 32   
  !
  type, bind(c) :: julianday
    integer(c_int64_t) :: day  !> the actual Julian day
    integer(c_int64_t) :: ms   !> the milisecond on that particular day
  end type julianday
  !
  interface
     function my_newjulianday(day, ms) result(c_pointer) bind(c, name='newJulianDay')
      import :: c_int64_t, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: day
      integer(c_int64_t), value :: ms
    end function my_newjulianday
    !
    subroutine my_deallocatejulianday(jd) bind(c,name='deallocateJulianDay')
      import :: c_ptr
      type(c_ptr), value :: jd
    end subroutine my_deallocatejulianday
    !
    function my_juliandaytostring(my_julianday, string) result(string_ptr) bind(c, name='juliandayToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_julianday
      character(c_char), dimension(*) :: string
    end function my_juliandaytostring
    !
  end interface
  !
contains
  !
  !>
  !! @brief construct a new Julian date 
  !!
  !! @param[in] day            the Julian day
  !! @param[in] ms             an integer denoting the actual milli seconds of a day 
  !! @return    ret_julianday  a pointer of type(julianday)
  function newJulianday(day, ms, errno) result(ret_julianday) !OK-TESTED.
    type(julianday), pointer :: ret_julianday
    integer(c_int64_t), intent(in) :: day
    integer(c_int64_t), intent(in) :: ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newjulianday(day, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 1 * 100 + 1
    call c_f_pointer(c_pointer, ret_julianday)
  end function newJulianday
  !
  !>
  !! @brief destructor for a Julian date 
  !!
  !! @param     my_julianday   a pointer of type(julianday)
  subroutine deallocateJulianday(my_julianday) !OK-TESTED.
    type(julianday), pointer :: my_julianday
    call my_deallocatejulianday(c_loc(my_julianday))
    my_julianday => null()
  end subroutine deallocateJulianday
  !
  !>
  !! @brief get Julian day as a string.
  !!
  !! @param[in]  my_julianday   a pointer to type(julianday). The Julian day to be converted to a string 
  !! @param[out] string         the Julian day verbose
  subroutine juliandayToString(my_julianday, string, errno) !OK-TESTED.
    type(julianday), pointer :: my_julianday
    character(len=max_julianday_str_len), intent(out) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_juliandaytostring(c_loc(my_julianday), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 1 * 100 + 3
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine juliandayToString
  !
end module mtime_julianday
!>
!! @brief Date and some operations supported on Date.
!!
!! @details
!!
!___________________________________________________________________________________________________________
module mtime_date
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_char, c_ptr, c_null_char, &
                                       & c_loc, c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_date_str_len
  public :: date
  public :: newdate
  public :: deallocateDate
  public :: replaceDate
  public :: dateToString
  public :: dateToPosixString
  !
  !> provides a string length for toString 
  integer, parameter :: max_date_str_len = 32
  !
  type, bind(c) :: date
    integer(c_int64_t) :: year
    integer(c_int) :: month
    integer(c_int) :: day
  end type date
  !
  interface newdate
    module procedure newdatefromstring
    module procedure newdatefromraw
    module procedure newdatefromraw_yi8
    module procedure newdatefromconstructandcopy
  end interface newdate
  !
  interface
    !
    function my_newdatefromstring(string) result(c_pointer) bind(c, name='newDate')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newdatefromstring
    !
    function my_newrawdate(year, month, day) result(c_pointer) bind(c, name='newRawDate')
      import :: c_int64_t, c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day
    end function my_newrawdate
    !
    function my_constructandcopydate(d) result(c_pointer) bind(c,name='constructAndCopyDate')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: d
    end function my_constructandcopydate
    !
    subroutine my_deallocatedate(d) bind(c, name='deallocateDate')
      import :: c_ptr
      type(c_ptr), value :: d
    end subroutine my_deallocatedate
    !
    function my_replacedate(src, dest) result(ret_dest) bind(c, name='replaceDate')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src , dest
    end function my_replacedate
    !
    function my_datetostring(my_date, string) result(string_ptr) bind(c, name='dateToString') 
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_date
      character(c_char), dimension(*) :: string
    end function my_datetostring
    !
    function my_datetoposixstring(my_date, string, fmtstr) result(string_ptr) bind(c, name='dateToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_date
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_datetoposixstring
    !
  end interface
  !
contains
  !
  !>
  !! @brief construct a new date 
  !!
  !! @param[in] string         an ISO 8601 conforming date string
  !! @return    ret_date       a pointer of type(date)
  function newdatefromstring(string,errno) result(ret_date)  !OK-TESTED.
    type(date), pointer :: ret_date
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional :: errno
    if (present(errno)) errno = 0
    c_pointer = my_newdatefromstring(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 1
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromstring
  !TODO: Doxy comments.
  function newdatefromraw_yi8(year, month, day, errno) result(ret_date) !OK-TESTED.
    type(date), pointer :: ret_date
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdate(year, month, day)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 2
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromraw_yi8
  !TODO: Doxy comments.
  function newdatefromraw(year, month, day, errno) result(ret_date) !OK-TESTED.
    type(date), pointer :: ret_date
    integer(c_int), intent(in) :: year, month, day
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdate(int(year,c_int64_t), month, day)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 3
    call c_f_pointer(c_pointer, ret_date)
  end function newdatefromraw
  !TODO: Doxy comments..
  function newdatefromconstructandcopy(src, errno) result(dest) !OK-TESTED
    type(date), pointer :: dest
    type(date), pointer :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopydate(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 2 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newdatefromconstructandcopy
  !
  !! @brief destructor for a date 
  !!
  !! @param[in] my_date        a pointer of type(date)
  subroutine deallocateDate(my_date) !OK-TESTED.
    type(date), pointer :: my_date
    call my_deallocatedate(c_loc(my_date))
    my_date => null()
  end subroutine deallocateDate
  !
  subroutine replaceDate(dest, src, errno)  !OK-TESTED.
    type(date), target, intent(inout) :: dest
    type(date), target, intent(in) :: src
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_replacedate(c_loc(src), c_loc(dest))
    if ((.not. c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 6
  end subroutine replaceDate
  !
  subroutine dateToString(my_date, string, errno) !OK-TESTED.
    type(date), pointer :: my_date
    character(len=max_date_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetostring(c_loc(my_date), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine dateToString
  !
  subroutine dateToPosixString(my_date, string, fmtstr, errno) !OK-TESTED.
    type(date), pointer :: my_date
    character(len=max_date_str_len) :: string
    character(len=32) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetoposixstring(c_loc(my_date), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 2 * 100 + 8
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine dateToPosixString
  !
end module mtime_date
!>
!! @brief Time and some operations supported on Time
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_time
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_ptr, c_null_char, &
                                       & c_loc, c_f_pointer, c_associated
  !
  implicit none
  !
  private
  !
  public :: max_time_str_len
  public :: time
  public :: newtime
  public :: deallocateTime
  public :: replaceTime
  public :: timeToString
  public :: timeToPosixString
  !
  !> provides a string length for toString 
  integer, parameter :: max_time_str_len = 32
  !
  type, bind(c) :: time
    integer(c_int) :: hour
    integer(c_int) :: minute
    integer(c_int) :: second
    integer(c_int) :: ms
  end type time
  !
  interface newtime
    module procedure newtimefromstring
    module procedure newtimefromraw
    module procedure newtimefromconstructandcopy
  end interface newtime
  !
  interface
    !
    function my_newtimefromstring(string) result(c_pointer) bind(c, name='newTime')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newtimefromstring
    !
    function my_newrawtime(hour, minute, second, ms) result(c_pointer) bind(c, name='newRawTime')
      import :: c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int), value :: hour, minute, second, ms
    end function my_newrawtime
    !
    function my_constructandcopytime(t) result(c_pointer) bind(c,name='constructAndCopyTime')
    import :: c_ptr
    type(c_ptr) :: c_pointer
    type(c_ptr), value :: t
  end function my_constructandcopytime
    !
    subroutine my_deallocatetime(t) bind(c,name='deallocateTime') 
      import :: c_ptr
      type(c_ptr), value :: t
    end subroutine my_deallocatetime
    !
    function my_replacetime(src, dest) result(ret_dest) bind(c, name='replaceTime')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src , dest
    end function my_replacetime
    !
    function my_timetostring(my_time, string) result(string_ptr) bind(c, name='timeToString') 
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
    end function my_timetostring
    !
    function my_timetoposixstring(my_time, string, fmtstr) result(string_ptr) bind(c, name='timeToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_timetoposixstring
    !
  end interface
  !
contains
  !
  function newtimefromstring(string, errno) result(ret_time) !OK-TESTED.
    type(time), pointer :: ret_time
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newtimefromstring(trim(adjustl(string))//c_null_char)
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 1
    call c_f_pointer(c_pointer, ret_time)
  end function newtimefromstring
  !
  function newtimefromraw(hour, minute, second, ms, errno) result(ret_time) !OK-TESTED.
    type(time), pointer :: ret_time
    integer(c_int), intent(in) :: hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawtime(hour, minute, second, ms)
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 2
    call c_f_pointer(c_pointer, ret_time)
  end function newtimefromraw
  !
  function newtimefromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(time), pointer :: dest
    type(time), pointer :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopytime(c_loc(src))
    if ((.not.c_associated(c_pointer)) .and. present(errno)) errno = 3 * 100 + 3
    call c_f_pointer(c_pointer, dest)
  end function newtimefromconstructandcopy
  !
  subroutine deallocateTime(my_time) !OK-TESTED.
    type(time), pointer :: my_time
    call my_deallocatetime(c_loc(my_time))
    my_time => null()
  end subroutine deallocateTime
  !
  subroutine replaceTime(dest,src,errno) !OK-TESTED.
    type(time), target, intent(in) :: src
    type(time), target, intent(inout) :: dest
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_replacetime(c_loc(src), c_loc(dest))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 5
  end subroutine replaceTime
  !
  subroutine timeToString(my_time, string, errno) !OK-TESTED.
    type(time), pointer :: my_time
    character(len=max_time_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_timetostring(c_loc(my_time), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 6
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine timeToString
  !
  subroutine timeToPosixString(my_time, string, fmtstr, errno) !OK-TESTED.
    type(time), pointer :: my_time
    character(len=max_time_str_len) :: string
    character(len=32) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_timetoposixstring(c_loc(my_time), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 3 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine timeToPosixString
  !
end module mtime_time
!>
!! @brief DateTime and some operations supported on DateTime.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_datetime
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_char, c_ptr, c_null_char, &
                                       & c_loc, c_f_pointer, c_associated
  !
  use mtime_julianday
  use mtime_date
  use mtime_time
  !
  implicit none
  !
  private
  !
  public :: max_datetime_str_len
  public :: datetime
  public :: newDatetime
  public :: deallocateDatetime
  public :: datetimeToString
  public :: datetimeToPosixString
  public :: getNoOfDaysInMonthDateTime
  public :: getNoOfDaysInYearDateTime
  public :: getDayOfYearFromDateTime
  public :: getNoOfSecondsElapsedInMonthDateTime
  public :: getNoOfSecondsElapsedInDayDateTime
  public :: getJulianDayFromDatetime
  !
  public :: min, max
  public :: assignment(=)
  public :: operator(>)
  public :: operator(<)
  public :: operator(<=)
  public :: operator(>=)
  public :: operator(==)
  public :: operator(/=)
  !
  !> provides a string length for toString 
  integer, parameter :: max_datetime_str_len = 32
  !
  type, bind(c) :: datetime
    type(date) :: date
    type(time) :: time
  end type datetime

  !
  interface newDatetime
    module procedure newdatetimefromstring
    module procedure newdatetimefromraw
    module procedure newdatetimefromraw_yi8
    module procedure newdatetimefromconstructandcopy
  end interface newDatetime

  !
  interface min
     module procedure datetime_min
  end interface min
  !
  interface max
     module procedure datetime_max
  end interface max
  !
  interface assignment (=)
    module procedure replacedatetime
  end interface assignment (=)
  !
  interface operator (>)
    module procedure datetime_gt
  end interface operator (>)
  !
  interface operator (<)
    module procedure datetime_lt
  end interface operator (<)
  !
  interface operator (<=)
    module procedure datetime_lt_or_eq
  end interface operator (<=)
  !
  interface operator (>=)
    module procedure datetime_gt_or_eq
  end interface operator (>=)
  !
  interface operator (==)
    module procedure datetime_eq
  end interface operator (==)
  !
  interface operator (/=)
    module procedure datetime_ne
  end interface operator (/=)
  !
  interface
    !
    function my_newdatetime(string) result(c_pointer) bind(c, name='newDateTime')
      import :: c_ptr, c_char
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newdatetime
    !
    function my_newrawdatetime(year, month, day, hour, minute, second, ms) result(c_pointer) &
            &             bind(c, name='newRawDateTime')
      import :: c_int64_t, c_int, c_ptr
      type(c_ptr) :: c_pointer
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day, hour, minute, second, ms
    end function my_newrawdatetime
    !
    function my_constructandcopydatetime(dt) result(c_pointer) bind(c,name='constructAndCopyDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: dt
    end function my_constructandcopydatetime
    !
    subroutine my_deallocatedatetime(dt) bind(c,name='deallocateDateTime')
      import :: c_ptr
      type(c_ptr), value :: dt
    end subroutine my_deallocatedatetime
    !
    function my_comparedatetime(op1, op2) result(ret) bind(c, name='compareDatetime')
      import :: c_ptr, c_int
      integer(c_int) :: ret
      type(c_ptr), value :: op1
      type(c_ptr), value :: op2
    end function my_comparedatetime

    function my_replacedatetime(src, dest) result(ret_dest) bind(c, name='replaceDatetime')
      import :: c_ptr
      type(c_ptr) :: ret_dest
      type(c_ptr), value :: src
      type(c_ptr), value :: dest
    end function my_replacedatetime
    !
    function my_datetimetostring(my_time, string) result(string_ptr) bind(c, name='datetimeToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
    end function my_datetimetostring
    !
    function my_datetimetoposixstring(my_time, string, fmtstr) result(string_ptr) bind(c, name='datetimeToPosixString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_time
      character(c_char), dimension(*) :: string
      character(c_char), dimension(*) :: fmtstr
    end function my_datetimetoposixstring
    !
    function my_getnoofdaysinmonthdatetime(dt) bind(c, name='getNoOfDaysInMonthDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofdaysinmonthdatetime
      type(c_ptr), value :: dt
    end function my_getnoofdaysinmonthdatetime
    !
    function my_getnoofdaysinyeardatetime(dt) bind(c, name='getNoOfDaysInYearDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofdaysinyeardatetime
      type(c_ptr), value :: dt
    end function my_getnoofdaysinyeardatetime
    !
    function my_getdayofyearfromdatetime(dt) bind(c, name='getDayOfYearFromDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getdayofyearfromdatetime
      type(c_ptr), value :: dt
    end function my_getdayofyearfromdatetime
    !
    function my_getnoofsecondselapsedinmonthdatetime(dt) bind(c, name='getNoOfSecondsElapsedInMonthDateTime')
      import :: c_int64_t, c_ptr
      integer(c_int64_t) :: my_getnoofsecondselapsedinmonthdatetime
      type(c_ptr), value :: dt
    end function my_getnoofsecondselapsedinmonthdatetime
    !
    function my_getnoofsecondselapsedindaydatetime(dt) bind(c, name='getNoOfSecondsElapsedInDayDateTime')
      import :: c_int, c_ptr
      integer(c_int) :: my_getnoofsecondselapsedindaydatetime
      type(c_ptr), value :: dt
    end function my_getnoofsecondselapsedindaydatetime
    !
    function my_getjuliandayfromdatetime(dt, jd) bind(c, name='getJulianDayFromDateTime')
      import :: c_ptr
      type(c_ptr) :: my_getjuliandayfromdatetime
      type(c_ptr), value :: dt, jd
    end function my_getjuliandayfromdatetime
    !
  end interface
  !
contains
  !
  function newdatetimefromstring(string, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newdatetime(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 1
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromstring
  !
  function newdatetimefromraw_yi8(year, month, day, hour, minute, second, ms, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdatetime(year, month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 2
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromraw_yi8
  !
  function newdatetimefromraw(year, month, day, hour, minute, second, ms, errno) result(ret_datetime) !OK-TESTED
    type(datetime), pointer :: ret_datetime
    integer(c_int), intent(in) :: year, month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawdatetime(int(year,c_int64_t), month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 3
    call c_f_pointer(c_pointer, ret_datetime)
  end function newdatetimefromraw
  !
  function newdatetimefromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(datetime), pointer :: dest
    type(datetime), pointer :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopydatetime(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 4 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newdatetimefromconstructandcopy

  ! @return Minimum of two date time. In case of equality we return @p a.
  !
  ! @note This function does not return a copy of one of the arguments
  ! but returns only the corresponding result point to avoid
  ! unnecessary deallocate calls.
  function datetime_min(a,b) result(res)
    type(datetime), pointer :: res
    type(datetime), pointer :: a,b

    if (a > b) then
       res => b
    else
       res => a
    end if
  end function datetime_min

  ! @return Maximum of two date time. In case of equality we return @p a.
  !
  ! @note This function does not return a copy of one of the arguments
  ! but returns only the corresponding result point to avoid
  ! unnecessary deallocate calls.
  function datetime_max(a,b) result(res)
    type(datetime), pointer :: res
    type(datetime), pointer :: a,b

    if (a < b) then
       res => b
    else
       res => a
    end if
  end function datetime_max

  !
  subroutine deallocateDatetime(my_datetime) !OK-TESTED.
    type(datetime), pointer :: my_datetime
    call my_deallocatedatetime(c_loc(my_datetime))
    nullify(my_datetime)
  end subroutine deallocateDatetime
  !
  subroutine datetimeToString(my_datetime, string, errno) !OK-TESTED
    type(datetime), pointer :: my_datetime
    character(len=max_datetime_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetimetostring(c_loc(my_datetime), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 6
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine datetimeToString
  !
  subroutine datetimeToPosixString(my_datetime, string, fmtstr, errno) !OK-TESTED.
    type(datetime), pointer :: my_datetime
    character(len=max_datetime_str_len) :: string
    character(len=32) :: fmtstr
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_datetimetoposixstring(c_loc(my_datetime), string, fmtstr)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 7
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine datetimeToPosixString
  !
  ! NOTE: Do not call the function using replacedatetime(.,.) directly; Use overloaded '=' instead as in "dest = src"
  subroutine replacedatetime(dest, src) !OK-TESTED. 
    type(datetime), target, intent(inout) :: dest
    type(datetime), target, intent(in) :: src
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_replacedatetime(c_loc(src), c_loc(dest))
  end subroutine replacedatetime
  !
  function datetime_gt(op1, op2) result(gt) !OK-TESTED.
    logical :: gt
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == 1) then
      gt = .true.
    else
      gt = .false.
    endif
  end function datetime_gt
  !
  function datetime_lt(op1, op2) result(lt) !OK-TESTED.
    logical :: lt
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == -1) then
      lt = .true.
    else
      lt = .false.
    endif
  end function datetime_lt
  !
  function datetime_lt_or_eq(op1, op2) result(lt_or_eq) !OK-TESTED.
    logical :: lt_or_eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == -1)) then
      lt_or_eq = .true.
    else
      lt_or_eq = .false.
    endif
  end function datetime_lt_or_eq
  !
  function datetime_gt_or_eq(op1, op2) result(gt_or_eq) !OK-TESTED
    logical :: gt_or_eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == 1)) then
      gt_or_eq = .true.
    else
      gt_or_eq = .false.
    endif
  end function datetime_gt_or_eq
  !
  function datetime_eq(op1, op2) result(eq) !OK-TESTED.
    logical :: eq
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret == 0) then
      eq = .true.
    else
      eq = .false.
    endif
  end function datetime_eq
  !
  function datetime_ne(op1, op2) result(ne) !OK-TESTED.
    logical :: ne
    type(datetime), target, intent(in) :: op1
    type(datetime), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparedatetime(c_loc(op1), c_loc(op2))
    if (ret /= 0) then
      ne = .true.
    else
      ne = .false.
    endif
  end function datetime_ne
  !
  function getNoOfDaysInMonthDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfDaysInMonthDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfDaysInMonthDateTime = my_getnoofdaysinmonthdatetime(c_loc(dt))
    if (getNoOfDaysInMonthDateTime == 0 .and. present(errno)) errno = 4 * 100 + 15
  end function getNoOfDaysInMonthDateTime
  !
  function getNoOfDaysInYearDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfDaysInYearDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfDaysInYearDateTime = my_getnoofdaysinyeardatetime(c_loc(dt))
    if (getNoOfDaysInYearDateTime == 0 .and. present(errno)) errno = 4 * 100 + 16
  end function getNoOfDaysInYearDateTime
  !
  function getDayOfYearFromDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getDayOfYearFromDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getDayOfYearFromDateTime = my_getdayofyearfromdatetime(c_loc(dt))
    if (getDayOfYearFromDateTime == 0 .and. present(errno)) errno = 4 * 100 + 17
  end function getDayOfYearFromDateTime
  !
  function getNoOfSecondsElapsedInMonthDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int64_t) :: getNoOfSecondsElapsedInMonthDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfSecondsElapsedInMonthDateTime = my_getnoofsecondselapsedinmonthdatetime(c_loc(dt))
    if (getNoOfSecondsElapsedInMonthDateTime == -1 .and. present(errno)) errno = 4 * 100 + 18
  end function getNoOfSecondsElapsedInMonthDateTime
  !
  function getNoOfSecondsElapsedInDayDateTime(dt, errno) !OK-TESTED.
    type(datetime), target :: dt
    integer(c_int) :: getNoOfSecondsElapsedInDayDateTime
    integer, optional:: errno
    if (present(errno)) errno = 0
    getNoOfSecondsElapsedInDayDateTime = my_getnoofsecondselapsedindaydatetime(c_loc(dt))
    if (getNoOfSecondsElapsedInDayDateTime == -1 .and. present(errno)) errno = 4 * 100 + 19
  end function getNoOfSecondsElapsedInDayDateTime
  !
  subroutine getJulianDayFromDatetime(dt, jd, errno) !OK-TESTED.
    type(datetime), target :: dt
    type(julianday), pointer :: jd
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_getjuliandayfromdatetime(c_loc(dt), c_loc(jd))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 4 * 100 + 20
  end subroutine getJulianDayFromDatetime
  !
end module mtime_datetime
!>
!! @brief TimeDelta and some operations supported on TimeDelta.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_timedelta
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int32_t, c_int64_t, c_float, c_double, c_char, &
                                       & c_ptr, c_null_char, c_loc, c_f_pointer, c_associated
  !
  use mtime_datetime
  use mtime_date 
  !use mtime_time
  !
  implicit none
  !
  private
  !
  public :: max_timedelta_str_len
  public :: timedelta
  public :: divisionquotienttimespan
  public :: newTimedelta
  public :: deallocateTimedelta
  public :: getTimeDeltaFromDate
  public :: getTimeDeltaFromDateTime
  public :: getTotalMilliSecondsTimeDelta
  public :: getTotalSecondsTimeDelta
  public :: timedeltaToString
  public :: getPTStringFromMS
  public :: getPTStringFromSeconds
  public :: getPTStringFromMinutes
  public :: getPTStringFromHours
  public :: elementwiseScalarMultiplyTimeDelta
  public :: elementwiseScalarMultiplyTimeDeltaInv
  public :: elementwiseScalarMultiplyTimeDelta_long
  public :: elementwiseScalarMultiplyTimeDeltaInv_long
  public :: elementwiseAddTimeDeltatoTimeDelta
  public :: divideTimeDeltaInSeconds
  public :: divideDatetimeDifferenceInSeconds  
  public :: operator(+)
  public :: operator(-)
  public :: operator(*)
  public :: operator(>)
  public :: operator(<)
  public :: operator(<=)
  public :: operator(>=)
  public :: operator(==)
  public :: operator(/=)
  !
  !> provides a string length for toString 
  integer, parameter :: max_timedelta_str_len = 32
  !
  type, bind(c) :: timedelta
    character(c_char) :: sign
    integer(c_int64_t) :: year
    integer(c_int) :: month
    integer(c_int) :: day
    integer(c_int) :: hour
    integer(c_int) :: minute
    integer(c_int) :: second
    integer(c_int) :: ms
  end type timedelta
  !
  type, bind(c) :: divisionquotienttimespan
    integer(c_int64_t) :: quotient;
    integer(c_int64_t) :: remainder_in_ms;
  end type divisionquotienttimespan
  !
  interface newTimedelta
    module procedure newtimedeltafromstring
    module procedure newtimedeltafromraw
    module procedure newtimedeltafromraw_yi8
    module procedure newtimedeltafromconstructandcopy
  end interface newTimedelta
  !
  interface operator (+)
    module procedure addtimedeltatodatetime
    module procedure adddatetimetotimedelta
    module procedure addtimedeltatodate
    module procedure adddatetotimedelta
    module procedure elementwiseAddTimeDeltatoTimeDelta
  end interface operator (+)
  !
  interface operator (-)
    module procedure getTimeDeltaFromDate
    module procedure getTimeDeltaFromDateTime
  end interface operator (-)
  !
  interface operator (*)
    module procedure elementwiseScalarMultiplyTimeDelta
    module procedure elementwiseScalarMultiplyTimeDeltaInv
    module procedure elementwiseScalarMultiplyTimeDelta_long
    module procedure elementwiseScalarMultiplyTimeDeltaInv_long
    module procedure elementwiseScalarMultiplyTimeDelta_real
  end interface operator (*)
  !
  interface operator (>)
  module procedure timedelta_gt
  end interface operator (>)
  !
  interface operator (<)
  module procedure timedelta_lt
  end interface operator (<)
  !
  interface operator (<=)
  module procedure timedelta_lt_or_eq
  end interface operator (<=)
  !
  interface operator (>=)
  module procedure timedelta_gt_or_eq
  end interface operator (>=)
  !
  interface operator (==)
  module procedure timedelta_eq
  end interface operator (==)
  !
  interface operator (/=)
  module procedure timedelta_ne
  end interface operator (/=)
  !
  interface getPTStringFromSeconds
    module procedure getPTStringFromSecondsInt
    module procedure getPTStringFromSecondsFloat
    module procedure getPTStringFromSecondsDouble
  end interface getPTStringFromSeconds
  !
  interface
    !
    function my_newtimedeltafromstring(string) result(c_pointer) bind(c, name='newTimeDelta')
      import :: c_char, c_ptr
      type(c_ptr)                     :: c_pointer
      character(c_char), dimension(*) :: string
    end function my_newtimedeltafromstring
    !
    function my_newrawtimedelta(sign, year, month, day, hour, minute, second, ms) result(c_pointer) &
            &             bind(c, name='newRawTimeDelta')
      import :: c_int64_t, c_char, c_int, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), value :: sign
      integer(c_int64_t), value :: year
      integer(c_int), value :: month, day, hour, minute, second, ms
    end function my_newrawtimedelta
    !
    function my_constructandcopytimedelta(td) result(c_pointer) bind(c,name='constructAndCopyTimeDelta')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: td
    end function my_constructandcopytimedelta
    !
    subroutine my_deallocatetimedelta(dt) bind(c, name='deallocateTimeDelta')
      import :: c_ptr
      type(c_ptr), value :: dt
    end subroutine my_deallocatetimedelta
    !
    function my_comparetimedelta(op1, op2) result(ret) bind(c, name='compareTimeDelta')
      import :: c_ptr, c_int
      integer(c_int) :: ret
      type(c_ptr), value :: op1
      type(c_ptr), value :: op2
    end function my_comparetimedelta
    !
    function my_gettimedeltafromdate(my_date1,my_date2,timedelta_return) result(timedelta_ptr) &
             &    bind(c,name='getTimeDeltaFromDate')
      import :: c_ptr
      type(c_ptr) :: timedelta_ptr
      type(c_ptr), value :: my_date1, my_date2, timedelta_return
    end function my_gettimedeltafromdate
    !
    function my_gettimedeltafromdatetime(my_datetime1,my_datetime2,timedelta_return) result(timedelta_ptr) &
             &    bind(c,name='getTimeDeltaFromDateTime')
      import :: c_ptr
      type(c_ptr) :: timedelta_ptr
      type(c_ptr), value :: my_datetime1, my_datetime2, timedelta_return
    end function my_gettimedeltafromdatetime
    !
    function my_gettotalmillisecondstimedelta(my_timedelta, my_datetime) &
            &   bind(c, name='getTotalMilliSecondsTimeDelta')
      import :: c_ptr, c_int64_t
      integer(c_int64_t) :: my_gettotalmillisecondstimedelta
      type(c_ptr), value :: my_timedelta, my_datetime
    end function my_gettotalmillisecondstimedelta
    !
    function my_gettotalsecondstimedelta(my_timedelta, my_datetime) bind(c, name='getTotalSecondsTimeDelta')
      import :: c_ptr, c_int64_t
      integer(c_int64_t) :: my_gettotalsecondstimedelta
      type(c_ptr), value :: my_timedelta, my_datetime
    end function my_gettotalsecondstimedelta
    !
    function my_timedeltatostring(td, tostring) result(string_ptr) bind(c, name='timedeltaToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: td
      character(c_char), dimension(*) :: tostring
    end function my_timedeltatostring
    !
    function my_addtimedeltatodatetime(my_datetime, my_timedelta, ret_datetime) result(datetime_ptr) &
         &   bind(c, name='addTimeDeltaToDateTime')
      import :: c_ptr
      type(c_ptr) :: datetime_ptr
      type(c_ptr), value :: my_datetime, my_timedelta, ret_datetime
    end function my_addtimedeltatodatetime
    !
    function my_addtimedeltatodate(my_date, my_timedelta, ret_date) result(date_ptr) &
         &   bind(c, name='addTimeDeltaToDate')
      import :: c_ptr
      type(c_ptr) :: date_ptr
      type(c_ptr), value :: my_date, my_timedelta, ret_date
    end function my_addtimedeltatodate
    !
    function my_elementwisescalarmultiplytimedelta(my_timedelta, lambda, scaled_timedelta) result(timedelta_return) &
         &   bind(c, name='elementwiseScalarMultiplyTimeDelta')
      import :: c_ptr, c_int64_t
      type(c_ptr) ::timedelta_return
      type(c_ptr), value :: my_timedelta, scaled_timedelta
      integer(c_int64_t), value :: lambda
    end function my_elementwisescalarmultiplytimedelta
    !
    FUNCTION my_elementwisescalarmultiplytimedeltadp(my_timedelta, lambda, scaled_timedelta) RESULT(timedelta_return) &
         &   bind(c, name='elementwiseScalarMultiplyTimeDeltaDP')
      import :: c_ptr, c_double
      type(c_ptr) ::timedelta_return
      type(c_ptr), value :: my_timedelta, scaled_timedelta
      REAL(c_double), VALUE :: lambda
    END FUNCTION my_elementwisescalarmultiplytimedeltadp
    !
    function my_elementwiseAddTimeDeltatoTimeDelta(my_timedelta1,my_timedelta2,added_timedelta) result(timedelta_return) &
         &   bind(c, name='elementwiseAddTimeDeltatoTimeDelta')
      import :: c_ptr
      type(c_ptr) :: timedelta_return
      type(c_ptr), value :: my_timedelta1,my_timedelta2,added_timedelta
    end function my_elementwiseAddTimeDeltatoTimeDelta
    !
    function my_getptstringfromms(ms, tostring) result(string_ptr) bind(c, name='getPTStringFromMS')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t), value :: ms
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromms
    !
    function my_getptstringfromsecondsint(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSeconds')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsint
    !
    function my_getptstringfromsecondsfloat(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSecondsFloat')
      import :: c_ptr, c_float, c_char
      type(c_ptr) :: string_ptr
      real(c_float), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsfloat
    !
    function my_getptstringfromsecondsdouble(s, tostring) result(string_ptr) bind(c, name='getPTStringFromSecondsDouble')
      import :: c_ptr, c_double, c_char
      type(c_ptr) :: string_ptr
      real(c_double), value :: s
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromsecondsdouble
    !
    function my_getptstringfromminutes(m, tostring) result(string_ptr) bind(c, name='getPTStringFromMinutes')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t),value :: m
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromminutes
    !
    function my_getptstringfromhours(h, tostring) result(string_ptr) bind(c, name='getPTStringFromHours')
      import :: c_ptr, c_int64_t, c_char
      type(c_ptr) :: string_ptr
      integer(c_int64_t),value :: h
      character(c_char), dimension(*) :: tostring
    end function my_getptstringfromhours
    !
    function my_divideTimeDeltaInSeconds(dividend, divisor, quotient) result(ret_quotient) bind(c,name='divideTimeDeltaInSeconds') 
      import :: c_ptr
      type(c_ptr) :: ret_quotient
      type(c_ptr), value :: dividend 
      type(c_ptr), value :: divisor
      type(c_ptr), value :: quotient
    end function my_divideTimeDeltaInSeconds
    !
    function my_divideDatetimeDifferenceInSeconds(dt1, dt2, divisor, quotient) result(ret_quotient) &
         &                                                                     bind(c,name='divideDatetimeDifferenceInSeconds') 
      import :: c_ptr
      type(c_ptr) :: ret_quotient
      type(c_ptr), value :: dt1
      type(c_ptr), value :: dt2
      type(c_ptr), value :: divisor
      type(c_ptr), value :: quotient
    end function my_divideDatetimeDifferenceInSeconds
    !
  end interface
  !
contains
  !
  function newtimedeltafromstring(string, errno) result(ret_timedelta) !OK-TESTED.
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: string
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    ret_timedelta => null()
    c_pointer = my_newtimedeltafromstring(trim(adjustl(string))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) then
      errno = 5 * 100 + 1
    else
      call c_f_pointer(c_pointer, ret_timedelta)
    endif
  end function newtimedeltafromstring
  !
  function newtimedeltafromraw(sign, year, month, day, hour, minute, second, ms, errno) result(ret_timedelta) 
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: sign
    integer(c_int), intent(in) :: year, month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawtimedelta(sign, int(year,c_int64_t), month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 2
    call c_f_pointer(c_pointer, ret_timedelta)
  end function newtimedeltafromraw
  !
  function newtimedeltafromraw_yi8(sign, year, month, day, hour, minute, second, ms, errno) result(ret_timedelta)
    type(timedelta), pointer :: ret_timedelta
    character(len=*), intent(in) :: sign
    integer(c_int64_t), intent(in) :: year
    integer(c_int), intent(in) :: month, day, hour, minute, second, ms
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_newrawtimedelta(sign, year, month, day, hour, minute, second, ms)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 3
    call c_f_pointer(c_pointer, ret_timedelta)
  end function newtimedeltafromraw_yi8
  !
  function newtimedeltafromconstructandcopy(src, errno) result(dest) !OK-TESTED.
    type(timedelta), pointer :: dest
    type(timedelta), pointer :: src
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_constructandcopytimedelta(c_loc(src))
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 5 * 100 + 4
    call c_f_pointer(c_pointer, dest)
  end function newtimedeltafromconstructandcopy
  !
  subroutine deallocateTimedelta(my_timedelta) !OK-TESTED.
    type(timedelta), pointer :: my_timedelta
    call my_deallocatetimedelta(c_loc(my_timedelta))
    nullify(my_timedelta)
  end subroutine deallocateTimedelta
  !
  function timedelta_gt(op1, op2) result(gt)
    logical :: gt
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == 1) then
      gt = .true.
    else
      gt = .false.
    endif
  end function timedelta_gt
  !
  function timedelta_lt(op1, op2) result(lt)
    logical :: lt
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == -1) then
      lt = .true.
    else
      lt = .false.
    endif
  end function timedelta_lt
  !
  function timedelta_lt_or_eq(op1, op2) result(lt_or_eq)
    logical :: lt_or_eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == -1)) then
      lt_or_eq = .true.
    else
      lt_or_eq = .false.
    endif
  end function timedelta_lt_or_eq
  !
  function timedelta_gt_or_eq(op1, op2) result(gt_or_eq)
    logical :: gt_or_eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if ((ret == 0) .or. (ret == 1)) then
      gt_or_eq = .true.
    else
      gt_or_eq = .false.
    endif
  end function timedelta_gt_or_eq
  !
  function timedelta_eq(op1, op2) result(eq)
    logical :: eq
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret == 0) then
      eq = .true.
    else
      eq = .false.
    endif
  end function timedelta_eq
  !
  function timedelta_ne(op1, op2) result(ne)
    logical :: ne
    type(timedelta), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    integer(c_int) :: ret
    ret = my_comparetimedelta(c_loc(op1), c_loc(op2))
    if (ret /= 0) then
      ne = .true.
    else
      ne = .false.
    endif
  end function timedelta_ne
  !
  function getTimeDeltaFromDate(op1,op2) result(ret) !OK-TESTED.
    type(timedelta), target :: ret
    type(date), target, intent(in)  :: op1, op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_gettimedeltafromdate(c_loc(op1),c_loc(op2),c_loc(ret))
  end function getTimeDeltaFromDate
  !
  function getTimeDeltaFromDateTime(op1,op2) result(ret) !OK-TESTED.
    type(timedelta), target :: ret
    type(datetime), target, intent(in) :: op1, op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_gettimedeltafromdatetime(c_loc(op1),c_loc(op2),c_loc(ret))
  end function getTimeDeltaFromDateTime
  ! 
  !WARNING: TD 0 is error. If you know your TD is 0, ignore the error flag.
  function getTotalMilliSecondsTimeDelta(td, dt, errno)  !OK-TESTED.
    integer(c_int64_t) :: getTotalMilliSecondsTimeDelta
    type(timedelta), target, intent(in):: td
    type(datetime), target :: dt
    integer, optional:: errno
    if (present(errno)) errno = 0
    getTotalMilliSecondsTimeDelta = my_gettotalmillisecondstimedelta(c_loc(td), c_loc(dt))
    if ((getTotalMilliSecondsTimeDelta == 0) .and. present(errno)) errno = 5 * 100 + 8
  end function getTotalMilliSecondsTimeDelta
  !
  !WARNING: TD 0 is error. If you know your TD is 0, ignore the error flag.
  function getTotalSecondsTimeDelta(td, dt, errno) !OK-TESTED.
    integer(c_int64_t) :: getTotalSecondsTimeDelta
    type(timedelta), pointer :: td
    type(datetime), pointer :: dt
    integer, optional:: errno
    if (present(errno)) errno = 0
    getTotalSecondsTimeDelta = my_gettotalsecondstimedelta(c_loc(td), c_loc(dt))
    if ((getTotalSecondsTimeDelta == 0) .and. present(errno)) errno = 5 * 100 + 9
  end function getTotalSecondsTimeDelta
  !
  subroutine timedeltaToString(my_timedelta, string, errno) !OK-TESTED.
    type(timedelta), pointer :: my_timedelta
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    string(:) = ''
    dummy_ptr = my_timedeltatostring(c_loc(my_timedelta), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 5 * 100 + 10
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine timedeltaToString
  !
  function addTimedeltaToDatetime(op1, op2) result(ret) !OK-TESTED.
    type(datetime), target :: ret
    type(datetime), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodatetime(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addTimedeltaToDatetime
  !
  function addDatetimeToTimedelta(op2, op1) result(ret) !OK-TESTED.
    type(datetime), target :: ret
    type(datetime), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodatetime(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addDatetimeToTimedelta
  !
  function addTimedeltaToDate(op1, op2) result(ret) !OK-TESTED.
    type(date), target :: ret
    type(date), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodate(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addTimedeltaToDate
  !
  function addDateToTimedelta(op2, op1) result(ret) !OK-TESTED.
    type(date), target :: ret
    type(date), target, intent(in) :: op1
    type(timedelta), target, intent(in) :: op2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_addtimedeltatodate(c_loc(op1), c_loc(op2), c_loc(ret))
  end function addDateToTimedelta  
  !
  function elementwiseScalarMultiplyTimeDelta(base_td, ilambda) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int32_t), intent(in) :: ilambda
    type(timedelta), target, intent(in) :: base_td
    integer(c_int64_t) :: lambda
    type(c_ptr) :: dummy_ptr
    lambda = int(ilambda, c_int64_t)
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDelta
  !
  function elementwiseScalarMultiplyTimeDeltaInv(ilambda, base_td) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int32_t), intent(in) :: ilambda
    type(timedelta), target, intent(in) :: base_td
    integer(c_int64_t) :: lambda
    type(c_ptr) :: dummy_ptr
    lambda = int(ilambda, c_int64_t)
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDeltaInv
  !
  function elementwiseScalarMultiplyTimeDelta_long(base_td, lambda) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int64_t), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDelta_long
  !
  function elementwiseScalarMultiplyTimeDeltaInv_long(lambda, base_td) result(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    integer(c_int64_t), intent(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedelta(c_loc(base_td), lambda, c_loc(scaled_td))
  end function elementwiseScalarMultiplyTimeDeltaInv_long
  !
  FUNCTION elementwiseScalarMultiplyTimeDelta_real(base_td, lambda) RESULT(scaled_td) !OK-TESTED.
    type(timedelta), target :: scaled_td
    REAL(c_double), INTENT(in) :: lambda
    type(timedelta), target, intent(in) :: base_td
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwisescalarmultiplytimedeltadp(c_loc(base_td), lambda, c_loc(scaled_td))
  END FUNCTION elementwiseScalarMultiplyTimeDelta_real
  !
  function elementwiseAddTimeDeltatoTimeDelta(td1, td2) result(added_td) !OK-TESTED.
    type(timedelta), target :: added_td
    type(timedelta), target, intent(in) :: td1, td2
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_elementwiseaddtimedeltatotimedelta(c_loc(td1), c_loc(td2), c_loc(added_td))
  end function elementwiseAddTimeDeltatoTimeDelta
  !
  subroutine getPTStringFromMS(ms, string) !OK-TESTED.
    integer(c_int64_t), intent(in) :: ms
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    string(1:len(string)) = ' '
    dummy_ptr = my_getptstringfromms(ms, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromMS 
  !
  subroutine getPTStringFromSecondsInt(s, string) !OK-TESTED.
    integer(c_int64_t) :: s
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    dummy_ptr = my_getptstringfromsecondsint(s, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromSecondsInt
  !
  subroutine getPTStringFromSecondsFloat(s, string) !OK-TESTED.
    real(c_float) :: s
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    dummy_ptr = my_getptstringfromsecondsfloat(s, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromSecondsFloat
  !
  subroutine getPTStringFromSecondsDouble(s, string) !OK-TESTED.
    real(c_double) :: s
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    dummy_ptr = my_getptstringfromsecondsdouble(s, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromSecondsDouble
  !
  subroutine getPTStringFromMinutes(m, string) !OK-TESTED.
    integer(c_int64_t) :: m
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    dummy_ptr = my_getptstringfromminutes(m, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromMinutes
  !
  subroutine getPTStringFromHours(h, string) !OK-TESTED.
    integer(c_int64_t)                   :: h
    character(len=max_timedelta_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    dummy_ptr = my_getptstringfromhours(h, string)
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getPTStringFromHours
  !
  subroutine divideTimeDeltaInSeconds(dividend, divisor, quotient)
    type(timedelta), target, intent(in) :: dividend
    type(timedelta), target, intent(in) :: divisor
    type(divisionquotienttimespan), target, intent(out) :: quotient
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_dividetimedeltainseconds(c_loc(dividend), c_loc(divisor), c_loc(quotient))
  end subroutine divideTimeDeltaInSeconds
  !
  subroutine divideDatetimeDifferenceInSeconds(dt1, dt2, divisor, quotient)
    type(datetime), target, intent(in) :: dt1
    type(datetime), target, intent(in) :: dt2
    type(timedelta), target, intent(in) :: divisor
    type(divisionquotienttimespan), target, intent(out) :: quotient
    type(c_ptr) :: dummy_ptr
    dummy_ptr = my_dividedatetimedifferenceinseconds(c_loc(dt1), c_loc(dt2), c_loc(divisor), c_loc(quotient))
  end subroutine divideDatetimeDifferenceInSeconds
  !
end module mtime_timedelta
!>
!! @brief Definition of the basic event type and its methods.
!!
!! @details
!!
!___________________________________________________________________________________________________________
module mtime_events
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_char, c_null_char, c_bool, c_ptr, &
                                       & c_loc, c_f_pointer, c_associated, c_null_ptr
  !
  use mtime_datetime
  use mtime_timedelta
  !
  implicit none
  !
  private
  !
  public :: max_eventname_str_len
  public :: event
  public :: newEvent
  public :: deallocateEvent
  public :: eventToString
  public :: isCurrentEventActive
  public :: iseventNextInNextDays
  public :: iseventNextInNextMonths
  public :: iseventNextInNextYears
  public :: getTriggerNextEventAtDateTime
  public :: getTriggeredPreviousEventAtDateTime
  public :: getEventId
  public :: getEventName
  public :: getEventReferenceDateTime
  public :: getEventFirstDateTime
  public :: getEventLastDateTime
  public :: getEventInterval
  public :: getNextEventIsFirst
  public :: getEventisFirstInDay
  public :: getEventisFirstInMonth
  public :: getEventisFirstInYear
  public :: getEventisLastInDay
  public :: getEventisLastInMonth
  public :: getEventisLastInYear
  !
  !> provides a string length for toString 
  integer, parameter :: max_eventname_str_len = 132
  !
  type, bind(c) :: event
    type(c_ptr) :: nextEventInGroup
    integer(c_int64_t) :: eventId
    type(c_ptr) :: eventName
    type(c_ptr) :: eventReferenceDatetime
    type(c_ptr) :: eventFirstDatetime
    type(c_ptr) :: eventLastDatetime
    type(c_ptr) :: eventInterval
    logical(c_bool) :: triggerCurrentEvent
    logical(c_bool) :: nextEventIsFirst
    logical(c_bool) :: eventisFirstInDay
    logical(c_bool) :: eventisFirstInMonth
    logical(c_bool) :: eventisFirstInYear
    logical(c_bool) :: eventisLastInDay
    logical(c_bool) :: eventisLastInMonth
    logical(c_bool) :: eventisLastInYear
    type(c_ptr) :: triggerNextEventDateTime
    type(c_ptr) :: triggeredPreviousEventDateTime
  end type event
  !
  interface newEvent
    module procedure newEventWithString
    module procedure newEventWithDataTypes
  end interface newEvent
  !
  interface
    !
    function my_newevent(name, referenceDate, firstdate, lastDate, interval, offset) result(c_pointer) &
         &              bind(c, name='newEvent')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
      character(c_char), dimension(*) :: referenceDate
      character(c_char), dimension(*) :: firstDate
      character(c_char), dimension(*) :: lastDate
      character(c_char), dimension(*) :: interval
      character(c_char), dimension(*) :: offset
    end function my_newevent
    !
    function my_neweventwithdatatypes(name, referenceDate, firstdate, lastDate, interval, offset) &
         &                           result(c_pointer) bind(c, name='newEventWithDataType')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
      type(c_ptr), value :: referenceDate
      type(c_ptr), value :: firstDate
      type(c_ptr), value :: lastDate
      type(c_ptr), value :: interval
      type(c_ptr), value :: offset
    end function my_neweventwithdatatypes
    !
    subroutine my_deallocateevent(ev) bind(c,name='deallocateEvent')
      import :: c_ptr
      type(c_ptr), value :: ev
    end subroutine my_deallocateevent
    !
    function my_eventtostring(my_event, string) result(string_ptr) bind(c, name='eventToString')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_event
      character(c_char), dimension(*) :: string
    end function my_eventtostring
    !
    function my_isCurrentEventActive(my_event, my_datetime, plus_slack, minus_slack) &
             &                      result(ret) bind(c, name='isCurrentEventActive')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_datetime
      type(c_ptr), value :: plus_slack
      type(c_ptr), value :: minus_slack
      logical(c_bool) :: ret
    end function my_isCurrentEventActive
    !
    function my_iseventnextinnextdays(my_event) result(ret) bind(c, name='iseventNextInNextDays')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextdays
    !
    function my_iseventnextinnextmonths(my_event) result(ret) bind(c, name='iseventNextInNextMonths')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextmonths
    !
    function my_iseventnextinnextyears(my_event) result(ret) bind(c, name='iseventNextInNextYears')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_iseventnextinnextyears
    !
    function my_gettriggernexteventatdatetime(my_event,my_currentdatetime,my_datetime) result(c_pointer) &
    & bind(c, name='getTriggerNextEventAtDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_currentdatetime
      type(c_ptr), value :: my_datetime
    end function my_gettriggernexteventatdatetime
    !
    function my_gettriggeredpreviouseventatdatetime(my_event,my_datetime) result(c_pointer) &
    & bind(c, name='getTriggeredPreviousEventAtDateTime')
      import :: c_ptr
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_datetime
    end function my_gettriggeredpreviouseventatdatetime
    !
    function my_geteventname(my_event, string) result(c_pointer) bind(c, name='getEventName')
      import :: c_ptr, c_char
      type(c_ptr) :: c_pointer
      type(c_ptr), value :: my_event
      character(c_char), dimension(*) :: string
    end function my_geteventname
    !
    function my_getnexteventisfirst(my_event) result(ret) bind(c, name='getNextEventIsFirst')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_getnexteventisfirst
    !
    function my_geteventisfirstinday(my_event) result(ret) bind(c, name='getEventisFirstInDay')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinday
    !
    function my_geteventisfirstinmonth(my_event) result(ret) bind(c, name='getEventisFirstInMonth')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinmonth
    !
    function my_geteventisfirstinyear(my_event) result(ret) bind(c, name='getEventisFirstInYear')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventisfirstinyear
    !
    function my_geteventislastinday(my_event) result(ret) bind(c, name='getEventisLastInDay')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinday
    !
    function my_geteventislastinmonth(my_event) result(ret) bind(c, name='getEventisLastInMonth')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinmonth
    !
    function my_geteventislastinyear(my_event) result(ret) bind(c, name='getEventisLastInYear')
      import :: c_bool, c_ptr
      type(c_ptr), value :: my_event
      logical(c_bool) :: ret
    end function my_geteventislastinyear
    !
  end interface
  !
contains
  !
  function newEventWithString(name, referenceDate, firstdate, lastDate, interval, offset, errno) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    character(len=*), intent(in)           :: name
    character(len=*), intent(in)           :: referenceDate
    character(len=*), intent(in)           :: firstDate
    character(len=*), intent(in)           :: lastDate
    character(len=*), intent(in)           :: interval
    character(len=*), intent(in), OPTIONAL :: offset
    character(len=32)                      :: zeroOffset = "PT00S" !Optional offset's proxy string.
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    if(present(offset)) then !offset is not null.
         c_pointer = my_newevent(trim(adjustl(name))//c_null_char,          &
         &                  trim(adjustl(referenceDate))//c_null_char,      &
         &                  trim(adjustl(firstDate))//c_null_char,          &
         &                  trim(adjustl(lastDate))//c_null_char,           &
         &                  trim(adjustl(interval))//c_null_char,           &
         &                  trim(adjustl(offset))//c_null_char)
    else                     !offset is null. Pass a 0 offset. 
         c_pointer = my_newevent(trim(adjustl(name))//c_null_char,    &
         &                  trim(adjustl(referenceDate))//c_null_char,      &
         &                  trim(adjustl(firstDate))//c_null_char,          &
         &                  trim(adjustl(lastDate))//c_null_char,           &
         &                  trim(adjustl(interval))//c_null_char,           &
         &                  trim(adjustl(zeroOffset))//c_null_char)
    end if
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 6 * 100 + 1
    call c_f_pointer(c_pointer, ret_event)
  end function newEventWithString
  !
  function newEventWithDataTypes(name, referenceDate, firstdate, lastDate, interval, offset, errno) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    character(len=*), intent(in) :: name
    type(datetime), pointer :: referenceDate
    type(datetime), pointer :: firstDate
    type(datetime), pointer :: lastDate
    type(timedelta), pointer :: interval
    type(timedelta), pointer, OPTIONAL :: offset
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    if (present(offset)) then
    c_pointer = my_neweventwithdatatypes(trim(adjustl(name))//c_null_char, &
         &                  c_loc(referenceDate), c_loc(firstDate), c_loc(lastDate), c_loc(interval), c_loc(offset))
    else
    c_pointer = my_neweventwithdatatypes(trim(adjustl(name))//c_null_char, &
         &                  c_loc(referenceDate), c_loc(firstDate), c_loc(lastDate), c_loc(interval), c_null_ptr)
    end if
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 6 * 100 + 1
    call c_f_pointer(c_pointer, ret_event)
  end function newEventWithDataTypes
  !
  ! WARNING: If my_event was added to a group, this should never be called; 
  ! use removeEventFromEventGroup instead. 
  subroutine deallocateEvent(my_event) 
    type(event), pointer :: my_event
    call my_deallocateevent(c_loc(my_event))
    my_event => null()
  end subroutine deallocateEvent
  !
  subroutine eventToString(my_event, string, errno) !TODO:C code still incomplete.
    type(event), pointer :: my_event
    character(len=max_eventname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_eventtostring(c_loc(my_event), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 3
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine eventToString
  !
  function isCurrentEventActive(my_event, my_datetime, plus_slack, minus_slack) result(ret)
    type(event), pointer     :: my_event
    type(datetime), pointer  :: my_datetime
    type(timedelta), pointer, OPTIONAL :: plus_slack
    type(timedelta), pointer, OPTIONAL :: minus_slack
    logical(c_bool) :: ret
    if (present(plus_slack) .and. present(minus_slack)) then 
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                   c_loc(plus_slack),  c_loc(minus_slack))

    else if (present(plus_slack) .and. .not.present(minus_slack)) then
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                           c_loc(plus_slack),  c_null_ptr)

    else if (.not.present(plus_slack) .and. present(minus_slack)) then
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                          c_null_ptr,  c_loc(minus_slack))
    
    else
      ret = my_isCurrentEventActive(c_loc(my_event), c_loc(my_datetime), &
            &                                   c_null_ptr, c_null_ptr)
    
    end if
  end function isCurrentEventActive
  !
  function iseventNextInNextDays(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextdays(c_loc(my_event))
  end function iseventNextInNextDays
  !
  function iseventNextInNextMonths(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextmonths(c_loc(my_event))
  end function iseventNextInNextMonths
  !
  function iseventNextInNextYears(my_event) result(ret)
    type(event), pointer :: my_event
    logical(c_bool) :: ret
    ret = my_iseventnextinnextyears(c_loc(my_event))
  end function iseventNextInNextYears
  !
  subroutine getTriggerNextEventAtDateTime(my_event,my_currentdatetime,my_datetime, errno)
    type(event), target, intent(in) :: my_event
    type(datetime), target, intent(in) :: my_currentdatetime
    type(datetime), target :: my_datetime
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_gettriggernexteventatdatetime(c_loc(my_event),c_loc(my_currentdatetime),c_loc(my_datetime))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 8
  end subroutine getTriggerNextEventAtDateTime
  !
  subroutine getTriggeredPreviousEventAtDateTime(my_event,my_datetime,errno)
    type(event), target, intent(in) :: my_event
    type(datetime), target :: my_datetime
    type(c_ptr) :: dummy_ptr
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_gettriggeredpreviouseventatdatetime(c_loc(my_event),c_loc(my_datetime))
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 9
  end subroutine getTriggeredPreviousEventAtDateTime
  !
  function getEventId(my_event) result(ret_evtid) !OK-TESTED.
    integer(c_int64_t) :: ret_evtid
    type(event), pointer :: my_event
    ret_evtid = my_event%eventId
  end function getEventId
  !
  subroutine getEventName(my_event, string, errno) !OK-TESTED.
    type(event), pointer :: my_event
    character(len=max_eventname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional :: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_geteventname(c_loc(my_event), string)
    if ((.not.c_associated(dummy_ptr)) .and. present(errno)) errno = 6 * 100 + 11
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getEventName
  !
  function getEventReferenceDateTime(my_event) result(ret_referenceDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_referenceDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventReferenceDatetime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_referenceDateTime)
    else
      ret_referenceDateTime => null()
    endif
  end function getEventReferenceDateTime
  !
  function getEventFirstDateTime(my_event) result(ret_eventFirstDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_eventFirstDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventFirstDateTime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventFirstDateTime)
    else
      ret_eventFirstDateTime => null()
    endif
  end function getEventFirstDateTime
  !
  function getEventLastDateTime(my_event) result(ret_eventLastDateTime) !OK-TESTED.
    type(datetime), pointer :: ret_eventLastDateTime
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventLastDateTime
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventLastDateTime)
    else
      ret_eventLastDateTime => null()
    endif
  end function getEventLastDateTime
  !
  function getEventInterval(my_event) result(ret_eventInterval) !OK-TESTED.
    type(timedelta), pointer :: ret_eventInterval
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%eventInterval
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_eventInterval)
    else
      ret_eventInterval => null()
    endif
  end function getEventInterval
  !
  function getNextEventIsFirst(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_getnexteventisfirst(c_loc(my_event))
  end function getNextEventIsFirst
  !
  function getEventisFirstInDay(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinday(c_loc(my_event))
  end function getEventisFirstInDay
  !
  function getEventisFirstInMonth(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinmonth(c_loc(my_event))
  end function getEventisFirstInMonth
  !
  function getEventisFirstInYear(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventisfirstinyear(c_loc(my_event))
  end function getEventisFirstInYear
  !
  function getEventisLastInDay(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinday(c_loc(my_event))
  end function getEventisLastInDay
  !
  function getEventisLastInMonth(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinmonth(c_loc(my_event))
  end function getEventisLastInMonth
  !
  function getEventisLastInYear(my_event) result(ret)
    type(event), target, intent(in) :: my_event
    logical(c_bool) :: ret
    ret = my_geteventislastinyear(c_loc(my_event))
  end function getEventisLastInYear
  !
end module mtime_events
!>
!! @brief Event-groups which contains a list of events.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_eventgroups
  !
  use, intrinsic :: iso_c_binding, only: c_int64_t, c_ptr, c_char, c_null_char, c_bool, &
                                       & c_loc, c_f_pointer, c_associated
  !
  use mtime_events
  !
  implicit none
  !
  private
  !
  public :: max_groupname_str_len
  public :: eventgroup
  public :: newEventGroup
  public :: deallocateEventGroup
  public :: addEventToEventGroup
  public :: removeEventFromEventGroup
  public :: getEventGroupId
  public :: getEventGroupName
  public :: getFirstEventFromEventGroup
  public :: getNextEventFromEventGroup
  !
  !> provides a string length for toString 
  integer, parameter :: max_groupname_str_len = 132
  !
  type, bind(c) :: eventgroup
    integer(c_int64_t) :: eventGroupId
    type(c_ptr) :: eventGroupName
    type(c_ptr) :: firstEventInGroup
  end type eventgroup
  !
  interface
    !
    function my_neweventgroup(name) result(c_pointer) bind(c, name='newEventGroup')
      import :: c_char, c_ptr
      type(c_ptr) :: c_pointer
      character(c_char), dimension(*) :: name
    end function my_neweventgroup
    !
    subroutine my_deallocateeventgroup(evgrp) bind(c,name='deallocateEventGroup')
      import :: c_ptr
      type(c_ptr), value :: evgrp
    end subroutine my_deallocateeventgroup
    !
    function my_addeventtoeventgroup(my_event, my_eventgroup) result(ret) bind(c, name='addNewEventToEventGroup')
      import :: c_bool, c_ptr
      logical(c_bool) :: ret
      type(c_ptr), value :: my_event
      type(c_ptr), value :: my_eventgroup
    end function my_addeventtoeventgroup
    !
    function my_removeeventfromeventgroup(evname, evgrp) result(ret) bind(c, name='removeEventFromEventGroup')
      import :: c_bool, c_char, c_ptr
      logical(c_bool) :: ret
      character(c_char), dimension(*) :: evname
      type(c_ptr), value :: evgrp
    end function my_removeeventfromeventgroup
    !
    function my_geteventgroupname(my_eventgroup, string) result(string_ptr) bind(c, name='getEventGroupName')
      import :: c_ptr, c_char
      type(c_ptr) :: string_ptr
      type(c_ptr), value :: my_eventgroup
      character(c_char), dimension(*) :: string
    end function my_geteventgroupname
    !
  end interface
  !
contains
  !
  function newEventGroup(name, errno) result(ret_eventgroup) !OK-TESTED.
    type(eventgroup), pointer :: ret_eventgroup 
    character(len=*), intent(in) :: name
    type(c_ptr) :: c_pointer
    integer, optional:: errno
    if (present(errno)) errno = 0
    c_pointer = my_neweventgroup(trim(adjustl(name))//c_null_char)
    if ((.not. c_associated(c_pointer)) .and. present(errno)) errno = 7 * 100 + 1
    call c_f_pointer(c_pointer, ret_eventgroup)
  end function newEventGroup
  !
  subroutine deallocateEventGroup(my_eventgroup) !OK-TESTED.
    type(eventgroup), pointer :: my_eventgroup
    call my_deallocateeventgroup(c_loc(my_eventgroup))
    my_eventgroup => null()
  end subroutine deallocateEventGroup
  !
  function addEventToEventGroup(my_event, my_eventGroup) result(ret) !OK-TESTED.
    logical :: ret
    type(event), pointer :: my_event
    type(eventgroup), pointer :: my_eventgroup
    ret = my_addeventtoeventgroup(c_loc(my_event), c_loc(my_eventgroup))
  end function addEventToEventGroup
  !
  function removeEventfromEventGroup(my_name, my_eventGroup) result(ret) !OK-TESTED.
    logical :: ret
    character(len=*), intent(in) :: my_name
    type(eventgroup), pointer :: my_eventgroup
    ret = my_removeeventfromeventgroup(trim(adjustl(my_name))//c_null_char, c_loc(my_eventgroup))
  end function removeEventFromEventGroup
  !
  function getEventGroupId(my_eventgroup) result(ret_grpid) !OK-TESTED.
    integer(c_int64_t) :: ret_grpid
    type(eventgroup), pointer :: my_eventgroup
    ret_grpid = my_eventgroup%eventGroupId
  end function getEventGroupId
  !
  subroutine getEventGroupName(my_eventgroup, string, errno)  !TESTED-OK.
    type(eventgroup), pointer :: my_eventgroup
    character(len=max_groupname_str_len) :: string
    type(c_ptr) :: dummy_ptr
    integer :: i
    integer, optional:: errno
    if (present(errno)) errno = 0
    dummy_ptr = my_geteventgroupname(c_loc(my_eventgroup), string)
    if ((.not. c_associated(dummy_ptr)) .and. present(errno)) errno = 7 * 100 + 6
    char_loop: do i = 1 , len(string)
      if (string(i:i) == c_null_char) exit char_loop
    end do char_loop
    string(i:len(string)) = ' '
  end subroutine getEventGroupName
  !
  function getFirstEventFromEventGroup(my_eventgroup) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    type(eventgroup), pointer :: my_eventgroup
    type(c_ptr) :: c_pointer
    c_pointer = my_eventgroup%firstEventInGroup
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_event)
    else
      ret_event => null()
    endif
  end function getFirstEventFromEventGroup
  !
  function getNextEventFromEventGroup(my_event) result(ret_event) !OK-TESTED.
    type(event), pointer :: ret_event
    type(event), pointer :: my_event
    type(c_ptr) :: c_pointer
    c_pointer = my_event%nextEventInGroup
    if (c_associated(c_pointer)) then
      call c_f_pointer(c_pointer, ret_event)
    else
      ret_event => null()
    endif
  end function getNextEventFromEventGroup
  !
end module mtime_eventgroups
!>
!! @brief Support for handling ISO 8601:2004 repeated time interval strings.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime_utilities
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_char, c_null_char
  !
  implicit none
  !
  private
  !
  public :: splitRepetitionString
  public :: getRepetitions
  public :: max_repetition_str_len
  !
  integer, parameter :: max_repetition_str_len = 132
  !
  interface
    !
    function my_getRepetitions(repetitionString) bind(c, name='getRepetitions')
      import :: c_int, c_char
      integer(c_int) :: my_getRepetitions
      character(c_char), dimension(*) :: repetitionString
    end function my_getRepetitions
    !
    subroutine my_splitRepetitionString(recurringTimeInterval, repetitor, start, end, duration) &
         bind(c, name='splitRepetitionString')
      import :: c_char
      character(c_char), dimension(*) :: recurringTimeInterval
      character(c_char), dimension(*) :: repetitor
      character(c_char), dimension(*) :: start
      character(c_char), dimension(*) :: end
      character(c_char), dimension(*) :: duration
    end subroutine my_splitRepetitionString
    !
  end interface
  !
contains
  !
  function getRepetitions(repetitionString) RESULT(r)
    integer :: r
    character(len=*), intent(in) :: repetitionString
    r = my_getRepetitions(trim(adjustl(repetitionString))//c_null_char)
  end function getRepetitions    
  !
  subroutine splitRepetitionString(recurringTimeInterval, repetitor, start, end, duration, lrepetitor, lstart, lend, lduration)
    character(len=*), intent(in) :: recurringTimeInterval
    character(len=*), intent(out) :: repetitor
    character(len=*), intent(out) :: start
    character(len=*), intent(out) :: end
    character(len=*), intent(out) :: duration
    
    logical, intent(out) :: lrepetitor
    logical, intent(out) :: lstart
    logical, intent(out) :: lend
    logical, intent(out) :: lduration
  
    integer :: i

    lrepetitor = .false.
    lstart = .false.
    lend = .false.
    lduration = .false.
    
    call my_splitRepetitionString(trim(adjustl(recurringTimeInterval))//c_null_char, repetitor, start, end, duration)
    
    if (repetitor(1:1) /= c_null_char) lrepetitor = .true. 
    char_loop1: do i = 1 , len(repetitor)
      if (repetitor(i:i) == c_null_char) exit char_loop1
    end do char_loop1
    repetitor(i:len(repetitor)) = ' '
    
    if (start(1:1) /= c_null_char) lstart = .true. 
    char_loop2: do i = 1 , len(start)
      if (start(i:i) == c_null_char) exit char_loop2
    end do char_loop2
    start(i:len(start)) = ' '
    
    if (end(1:1) /= c_null_char) lend = .true. 
    char_loop3: do i = 1 , len(end)
      if (end(i:i) == c_null_char) exit char_loop3
    end do char_loop3
    end(i:len(end)) = ' '
  
    if (duration(1:1) /= c_null_char) lduration = .true. 
    char_loop4: do i = 1 , len(duration)
      if (duration(i:i) == c_null_char) exit char_loop4
    end do char_loop4
    duration(i:len(duration)) = ' '
  
  end subroutine splitRepetitionString
  !
end module mtime_utilities
!>
!! @}
!>
!! @brief mtime is a compound module making all library components accessible via one module.
!!
!! @details
!!
!! @author  Luis Kornblueh, Max Planck Institute for Meteorology
!! @author  Rahul Sinha, Max Planck Institute for Meteorology
!!
!___________________________________________________________________________________________________________
module mtime
  !
  use, intrinsic :: iso_c_binding, only: c_int, c_int64_t, c_bool, c_ptr, c_char
  !
  use mtime_calendar
  use mtime_julianday
  use mtime_date
  use mtime_time
  use mtime_datetime
  use mtime_timedelta
  use mtime_events
  use mtime_eventgroups
  use mtime_utilities
  !
  implicit none
  !
  public
  !
  ! integer, parameter :: no_of_sec_in_a_day    = 86400
  ! integer, parameter :: no_of_sec_in_a_hour   =  3600
  ! integer, parameter :: no_of_sec_in_a_minute =    60
  ! !  
  ! integer, parameter :: no_of_ms_in_a_day     = 86400000
  ! integer, parameter :: no_of_ms_in_half_day  = 43200000
  ! integer, parameter :: no_of_ms_in_a_hour    =  3600000
  ! integer, parameter :: no_of_ms_in_a_minute  =    60000
  ! integer, parameter :: no_of_ms_in_a_second  =     1000
  !
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_DAY') :: no_of_sec_in_a_day
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_HOUR') :: no_of_sec_in_a_hour
  integer(c_int), bind(c,name='NO_OF_SEC_IN_A_MINUTE') :: no_of_sec_in_a_minute
  !  
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_DAY') :: no_of_ms_in_a_day
  integer(c_int), bind(c,name='NO_OF_MS_IN_HALF_DAY') :: no_of_ms_in_half_day
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_HOUR') :: no_of_ms_in_a_hour
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_MINUTE') :: no_of_ms_in_a_minute
  integer(c_int), bind(c,name='NO_OF_MS_IN_A_SECOND') :: no_of_ms_in_a_second
  ! 
  integer, parameter :: max_mtime_error_str_len = 132
  !
  integer, parameter :: no_error                                        = 0,          &

       &                calendar_calendartostring                       = 0*100 +  1, &
 
       &                julianday_newjulianday                          = 1*100 +  1, &
       &                julianday_juliandaytostring                     = 1*100 +  3, &
       
       &                date_newdatefromstring                          = 2*100 +  1, &
       &                date_newdatefromraw_yi8                         = 2*100 +  2, &
       &                date_newdatefromraw                             = 2*100 +  3, &
       &                date_newdatefromconstructandcopy                = 2*100 +  4, &
       &                date_replacedate                                = 2*100 +  6, &
       &                date_datetostring                               = 2*100 +  7, &
       &                date_datetoposixstring                          = 2*100 +  8, &
       
       &                time_newtimefromstring                          = 3*100 +  1, &
       &                time_newtimefromraw                             = 3*100 +  2, &
       &                time_newtimefromconstructandcopy                = 3*100 +  3, &
       &                time_replacetime                                = 3*100 +  5, &
       &                time_timetostring                               = 3*100 +  6, &
       &                time_timetoposixstring                          = 3*100 +  7, &
       
       &                datetime_newdatetimefromstring                  = 4*100 +  1, &
       &                datetime_newdatetimefromraw_yi8                 = 4*100 +  2, &
       &                datetime_newdatetimefromraw                     = 4*100 +  3, &
       &                datetime_newdatetimefromconstructandcopy        = 4*100 +  4, &
       &                datetime_datetimetostring                       = 4*100 +  6, &
       &                datetime_datetimetoposixstring                  = 4*100 +  7, &
       &                datetime_getnoofdaysinmonthdatetime             = 4*100 + 15, &
       &                datetime_getnoofdaysinyeardatetime              = 4*100 + 16, &
       &                datetime_getdayofyearfromdatetime               = 4*100 + 17, &
       &                datetime_getnoofsecondselapsedinmonthdatetime   = 4*100 + 18, &
       &                datetime_getnoofsecondselapsedindaydatetime     = 4*100 + 19, &
       &                datetime_getjuliandayfromdatetime               = 4*100 + 20, &
       
       &                timedelta_newtimedeltafromstring                = 5*100 +  1, &
       &                timedelta_newtimedeltafromraw                   = 5*100 +  2, &
       &                timedelta_newtimedeltafromraw_yi8               = 5*100 +  3, &
       &                timedelta_newtimedeltafromconstructandcopy      = 5*100 +  4, &
       &                timedelta_gettotalmillisecondstimedelta         = 5*100 +  8, &
       &                timedelta_gettotalsecondstimedelta              = 5*100 +  9, &
       &                timedelta_timedeltatostring                     = 5*100 + 10, &
       
       &                events_newevent                                 = 6*100 +  1, &
       &                events_eventtostring                            = 6*100 +  3, &
       &                events_gettriggernexteventatdatetime            = 6*100 +  8, &
       &                events_gettriggeredpreviouseventatdatetime      = 6*100 +  9, &
       &                events_geteventname                             = 6*100 + 11, &
       
       &                eventgroups_neweventgroup                       = 7*100 +  1, &
       &                eventgroups_geteventgroupname                   = 7*100 +  6
  !
contains
  !
  subroutine mtime_strerror(errno, error_message)
    integer, intent(in) :: errno
    character(len=*), intent(out) :: error_message

    select case (errno) 
    case (no_error) 
      error_message = 'no error'
      
    case (calendar_calendartostring)
      error_message = 'could not retrieve the string in <calendartostring>'
      
    case (julianday_newjulianday)
      error_message = 'could not allocate julianday in <newjulianday>'
    case (julianday_juliandaytostring)
      error_message = 'could not retrieve the string in <juliandaytostring>'
      
    case (date_newdatefromstring)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromraw_yi8)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromraw)
      error_message = 'could not allocate date in <newdate>'
    case (date_newdatefromconstructandcopy)
      error_message = 'could not allocate date in <newdate>'
    case (date_replacedate)
      error_message = 'replace-date failed in <replacedate>'
    case (date_datetostring)
      error_message = 'could not retrieve the string in <datetostring>'
    case (date_datetoposixstring)
      error_message = 'could not retrieve the string in <datetoposixstring>'

    case (time_newtimefromstring)
      error_message = 'could not allocate time in <newtime>'
    case (time_newtimefromraw)
      error_message = 'could not allocate time in <newtime>'
    case (time_newtimefromconstructandcopy)
      error_message = 'could not allocate time in <newtime>'
    case (time_replacetime)
      error_message = 'replace-time failed in <replacetime>'
    case (time_timetostring)
      error_message = 'could not retrieve the string in <timetostring>'
    case (time_timetoposixstring)
      error_message = 'could not retrieve the string in <timetoposixstring>'

    case (datetime_newdatetimefromstring)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromraw_yi8)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromraw)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_newdatetimefromconstructandcopy)
      error_message = 'could not allocate datetime in <newdatetime>'
    case (datetime_datetimetostring)
      error_message = 'could not retrieve the string in <timetostring>'
    case (datetime_datetimetoposixstring)
      error_message = 'could not retrieve the string in <timetoposixstring>'
    case (datetime_getnoofdaysinmonthdatetime)
      error_message = 'could not retrieve the no-of-days-in-month-datetime in <getnoofdaysinmonthdatetime>'
    case (datetime_getnoofdaysinyeardatetime)
      error_message = 'could not retrieve the no-of-days-in-year-datetime in <getnoofdaysinyeardatetime>'
    case (datetime_getdayofyearfromdatetime)
      error_message = 'could not calculate the day-of-year-from-date-time in <getdayofyearfromdatetime>'
    case (datetime_getnoofsecondselapsedinmonthdatetime)
      error_message = 'could not calculate the no-of-seconds-elapsed-in-month-datetime in <getnoofsecondselapsedinmonthdatetime>'
    case (datetime_getnoofsecondselapsedindaydatetime)
      error_message = 'could not calculate the no-of-seconds-elapsed-in-day-datetime in <getnoofsecondselapsedindaydatetime>'
    case (datetime_getjuliandayfromdatetime)
      error_message = 'could not retrieve the get-julian-day-from-datetime in <getjuliandayfromdatetime>'
      
    case (timedelta_newtimedeltafromstring)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromraw)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromraw_yi8)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_newtimedeltafromconstructandcopy)
      error_message = 'could not allocate timedelta in <newtimedelta>'
    case (timedelta_gettotalmillisecondstimedelta)
      error_message = 'error in calculating total-milli-seconds-timedelta in <gettotalmillisecondstimedelta>'
    case (timedelta_gettotalsecondstimedelta)
      error_message = 'error in calculating total-seconds-in-timedelta in <gettotalmillisecondstimedelta>'
    case (timedelta_timedeltatostring)
      error_message = 'could not retrieve the string in <timedeltatostring>'
      
    case (events_newevent)
      error_message = 'could not allocate event in <newevent>'
    case (events_eventtostring)
      error_message = 'could not retrieve the string in <eventtostring>'
    case (events_gettriggernexteventatdatetime)
      error_message = 'failed to retrieve trigger-next-event-at-datetime in <gettriggernexteventatdatetime>'
    case (events_gettriggeredpreviouseventatdatetime)
      error_message = 'failed to retrieve triggered-previous-event-at-date-time in <gettriggeredpreviouseventatdatetime>'
    case (events_geteventname)
      error_message = 'failed to retrieve event-name in <geteventname>'
      
    case (eventgroups_neweventgroup)
      error_message = 'could not allocate eventgroup in <neweventgroup>'
    case (eventgroups_geteventgroupname)
      error_message = 'could not retrieve the group-name in <geteventgroupname>'
    end select
  end subroutine mtime_strerror
  !
end module mtime
